"""Functions to write calculated outputs to a file."""

import datetime
import json
import pathlib
from typing import Optional

from mobi_motion_tracking.core import models


def generate_output_filename(
    gold_participant_ID: str, output_dir: pathlib.Path
) -> pathlib.Path:
    """Generates a unique filename based on gold participant ID and date.

    The filename follows the format: `results_<gold_participant_ID>_<MMDDYYYY>.ndjson`.
    If the directory does not exist, it is created. If the file does not exist, it is
    created.

    Args:
        gold_participant_ID: The identifier for the gold-standard participant.
        output_dir: The directory where the NDJSON file should be stored.

    Returns:
        pathlib.Path: The full path to the generated NDJSON file.
    """
    date_str = datetime.datetime.now().strftime("%m%d%Y")
    base_filename = f"results_{gold_participant_ID}_{date_str}.ndjson"
    output_dir.mkdir(parents=True, exist_ok=True)
    output_file = output_dir / base_filename

    if not output_file.exists():
        output_file.touch()

    return output_file


def save_results_to_ndjson(
    gold: models.ParticipantData,
    subject: models.ParticipantData,
    similarity_metrics: models.SimilarityMetrics,
    output_dir: pathlib.Path,
    selected_metrics: Optional[list[str]] = None,
) -> dict:
    """Appends results to a NDJSON file with selected or all similarity metrics.

    The function writes a new JSON object containing the subject's metadata,
    the method used, and the selected similarity metrics. If `selected_metrics`
    is not provided, all available metrics from `similarity_metrics.metrics` are
    written.

    The data is appended to an NDJSON file specific to the gold participant,
    with the filename generated using `generate_output_filename()`.

    Args:
        gold: data for the gold-standard participant.
        subject: data for the subject being analyzed.
        similarity_metrics: Object containing the similarity
            method and metrics.
        output_dir: Directory where the results file should be saved.
        selected_metrics: List of metric keys to include in the
            output. If None, all available metrics are written.

    Returns:
        dict: entry to be written to output file.

    Raises:
        ValueError: If any selected metric is not available in
            `similarity_metrics.metrics`.
    """
    new_entry = {
        "participant_ID": subject.participant_ID,
        "sheetname": subject.sequence_sheetname,
        "method": similarity_metrics.method,
    }

    if selected_metrics is None:
        selected_metrics = list(similarity_metrics.metrics.keys())

    for metric_key in selected_metrics:
        if metric_key in list(similarity_metrics.metrics.keys()):
            new_entry[metric_key] = similarity_metrics.metrics[metric_key]
        else:
            raise ValueError("Selected metrics are not eligible for selected method.")

    output_path = generate_output_filename(gold.participant_ID, output_dir)

    with open(output_path, "a") as f:
        json.dump(new_entry, f)
        f.write("\n")

    return new_entry



"""Performs preprocessing steps for raw data."""

import numpy as np

from mobi_motion_tracking.preprocessing.joint_index_list import DEFAULT_JOINT_SEGMENTS


def center_joints_to_hip(data: np.ndarray) -> np.ndarray:
    """Center all joints to the hip as origin.

    This function sets the coordinates of the hip (x,y,z) as a new
    relative origin (0,0,0) for each frame. The x,y,and z coordinates
    of the hip will be subtracted from the x, y, and z coordinates of
    all joints for every frame.

    Args:
        data: ndarray, cleaned raw data.

    Returns:
        normalized_data: ndarray, data normalized to the hip.
    """
    normalized_data = data.copy()
    x_pelvis = data[:, 1]
    y_pelvis = data[:, 2]
    z_pelvis = data[:, 3]

    normalized_data[:, 1::3] -= x_pelvis[:, np.newaxis]
    normalized_data[:, 2::3] -= y_pelvis[:, np.newaxis]
    normalized_data[:, 3::3] -= z_pelvis[:, np.newaxis]

    return normalized_data


def get_average_length(
    centered_data: np.ndarray, segment_list: list = DEFAULT_JOINT_SEGMENTS
) -> np.ndarray:
    """Calculate the average lengths of all joint segments.

    This function calculates the average length across all frames of all connecting
    joint segments in the skeleton. The x, y,and z coordinates of specified starting
    and ending joints based on a user provided segment_list or default_joint_segments
    in JOINT_INDEX_LIST are used to calculate the average distance between the two
    joints for all frames.

    Args:
        centered_data: centered data output from center_joints_to_hip. The
            first column in centered data contains frame number, the following columns
            contain joint coordinates.
        segment_list: List containing all coordinate index pairs for all joint segments
            in skeleton whose lengths will be normalized. Defaults to JOINT_INDEX_LIST.

    Returns:
        ndarray [N,1], average distance between joints for all segments.

    Raises:
        IndexError: when a joint index in JOINT_INDEX_LIST is out of range of total
            number of joints.
    """
    num_segments = len(segment_list)
    num_joint_coordinates = centered_data.shape[1]
    all_distances = np.zeros((centered_data.shape[0], num_segments))

    if np.any(np.array(segment_list) >= num_joint_coordinates):
        raise IndexError(
            "Incorrect joint index list. Joint index in \
                         segment_list is out of range for data."
        )

    for i, segment in enumerate(segment_list):
        start_indices = np.array([segment[0][0], segment[1][0], segment[2][0]])
        end_indices = np.array([segment[0][1], segment[1][1], segment[2][1]])

        start_points = centered_data[:, start_indices]
        end_points = centered_data[:, end_indices]

        distances = np.linalg.norm(start_points - end_points, axis=1)
        all_distances[:, i] = distances

    average_distances = all_distances.mean(axis=0, keepdims=True).T

    return average_distances


def normalize_segments(
    centered_data: np.ndarray,
    average_lengths: np.ndarray,
    segment_list: list = DEFAULT_JOINT_SEGMENTS,
) -> np.ndarray:
    """Normalize skeleton segments to maintain consistent bone lengths across frames.

    This function processes a motion sequence by adjusting the length of each skeleton
    segment to match the target lengths of the gold standard, while preserving the
    starting joint positions. The x, y, and z coordinates of specified starting
    and ending joints based on a user provided segment_list or default_joint_segments
    in JOINT_INDEX_LIST are used.

    Args:
        centered_data: centered data output from center_joints_to_hip. The
            first column in centered data contains frame number, the following columns
            contain joint coordinates.
        average_lengths: Array of shape (len(segment_list), 1) containing target
            lengths for each skeleton segment.
        segment_list: List containing all coordinate index pairs for all joint segments
            in skeleton whose lengths will be normalized. Defaults to JOINT_INDEX_LIST.

    Returns:
        np.ndarray: Normalized motion data with consistent bone lengths, same shape as
            centered_data.

    Raises:
        ValueError: when length of segment_list and length of average_lengths do not
            match.
        ValueError: when the number of columns in centered data does not correlate to
            the length of segment_list or length of average_lengths.
    """
    normalized_data = centered_data.copy()

    if len(segment_list) != len(average_lengths):
        raise ValueError("Mismatch in shape for segment_list and average_lengths.")

    if (centered_data.shape[1] != 3 * len(segment_list) + 4) or (
        centered_data.shape[1] != 3 * len(average_lengths) + 4
    ):
        raise ValueError(
            "The shape of centered_data does not match the expected dimensions."
        )

    for frame in range(centered_data.shape[0]):
        for i, segment in enumerate(segment_list):
            start_indices = np.array([segment[0][0], segment[1][0], segment[2][0]])
            end_indices = np.array([segment[0][1], segment[1][1], segment[2][1]])

            segment_vector = (
                centered_data[frame, end_indices] - centered_data[frame, start_indices]
            )

            scaled_segment_vector = (
                average_lengths[i] / np.linalg.norm(segment_vector)
            ) * segment_vector

            normalized_data[frame, end_indices] = (
                normalized_data[frame, start_indices] + scaled_segment_vector
            )
    return normalized_data



"""Python based runner."""

import pathlib
from typing import Literal

from mobi_motion_tracking.io.readers import readers
from mobi_motion_tracking.io.writers import writers
from mobi_motion_tracking.preprocessing import preprocessing
from mobi_motion_tracking.processing import similarity_functions

ALGORITHM_LIST = ["dtw"]


def run(
    experimental_path: pathlib.Path,
    gold_path: pathlib.Path,
    sequence: list[int],
    algorithm: Literal["dtw"] = "dtw",
) -> list:
    """Checks if experimental path is a directory or file, calls run_file.

    This function determines whether the experimental path is a directory or a single
    file and processes each subject's data accordingly by calling `run_file`.

    Args:
        experimental_path: Path to the subject's motion tracking data
            file or directory.
        gold_path: Path to the gold-standard motion tracking data file.
        sequence: List of sequence numbers to process.
        algorithm: Name of the algorithm to use for similarity computation.

    Returns:
        list of lists containing metadata and specified metrics for each
            subject.

    Raises:
        FileNotFoundError: Input 'experimental_path' doesn't exist.
        ValueError: if algorithm is unsupported.
        TypeError: If `experimental_path` is not a file or directory.
    """
    outputs = []

    if algorithm not in ALGORITHM_LIST:
        raise ValueError("Unsupported algorithm provided.")

    if experimental_path.is_dir():
        for file in experimental_path.iterdir():
            output_dir = experimental_path
            try:
                subject_output = run_file(
                    file, gold_path, output_dir, sequence, algorithm
                )
                outputs.append(subject_output)
            except ValueError as ve:
                print(f"Skipping file: {ve}")
    elif experimental_path.is_file():
        output_dir = experimental_path.parent
        subject_output = run_file(
            experimental_path, gold_path, output_dir, sequence, algorithm
        )
        outputs.append(subject_output)
    else:
        raise FileNotFoundError("Input path does not exist.")

    return outputs


def run_file(
    file_path: pathlib.Path,
    gold_path: pathlib.Path,
    output_dir: pathlib.Path,
    sequence: list[int],
    algorithm: Literal["dtw"] = "dtw",
) -> list:
    """Performs main processing steps for a subject, per sequence.

    This function reads motion tracking data from the specified subject and gold-
    standard files, applies preprocessing steps, computes similarity metrics using the
    specified algorithm, and saves the results. By default, the results are saved as a
    csv in the parent folder of experimental path. If a file passed in a directory is
    invalid, the function continues.

    Args:
        file_path: Path to the subject's motion tracking data file.
        gold_path: Path to the gold-standard motion tracking data file.
        output_dir: Directory where similarity results should be saved.
        sequence: List of sequence numbers to process.
        algorithm: Name of the algorithm to use for similarity computation.

    Returns:
        list of dictionaries being written to the output file.

    Raises:
        ValueError: Unsupported algorithm selected.
        ValueError: Invalid file extension.
        ValueError: Subject or gold file is named incorrectly.
    """
    if algorithm == "dtw":
        similarity_function = similarity_functions.dynamic_time_warping
    else:
        raise ValueError("Unsupported algorithm selected.")

    if ".xlsx" != file_path.suffix:
        raise ValueError(f"Invalid file extension: {file_path}. Expected '.xlsx'.")

    participant_ID = file_path.stem
    if not (participant_ID.isdigit() or "gold" in participant_ID.lower()):
        raise ValueError("The input file is named incorrectly.")

    results_list = []
    for seq in sequence:
        gold = readers.read_participant_data(gold_path, seq)
        subject = readers.read_participant_data(file_path, seq)

        if subject.data.size == 0:
            continue

        gold.data = preprocessing.center_joints_to_hip(gold.data)
        subject.data = preprocessing.center_joints_to_hip(subject.data)
        gold_average_lengths = preprocessing.get_average_length(gold.data)
        subject.data = preprocessing.normalize_segments(
            subject.data, gold_average_lengths
        )

        similarity_metric = similarity_function(gold.data, subject.data)

        results = writers.save_results_to_ndjson(
            gold,
            subject,
            similarity_metric,
            output_dir,
            selected_metrics=["distance"],
        )
        results_list.append(results)

    return results_list



"""Main function for mobi_motion_tracking."""

from mobi_motion_tracking.core import cli


def run_main() -> None:
    """Main entry point to mobi_motion_tracking."""
    cli.main()


if __name__ == "__main__":
    cli.main()



"""CLI for mobi-motion-tracking."""

import argparse
import pathlib
from typing import List, Optional

from mobi_motion_tracking.core import orchestrator


def parse_sequence_list(sequence_str: str) -> List[int]:
    """Converts input sequence string to List[int]."""
    return [int(seq.strip()) for seq in sequence_str.split(",")]


def parse_arguments(args: Optional[List[str]]) -> argparse.Namespace:
    """Argument parser for mobi-motion-tracking cli.

    Args:
        args: A list of command line arguments given as strings. If None, the parser
            will take the args from `sys.argv`.

    Returns:
        Namespace object with all the input arguments and default values.

    Raises:
        SystemExit: if arguments are None.
    """
    parser = argparse.ArgumentParser(
        description="Run the main motion tracking pipeline.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="Please report issues at https://github.com/childmindresearch/mobi-motion-tracking.",
    )
    parser.add_argument(
        "-d",
        "--data",
        type=pathlib.Path,
        required=True,
        help="Path to the subject(s) data.",
    )

    parser.add_argument(
        "-g",
        "--gold",
        type=pathlib.Path,
        required=True,
        help="Path to the gold data file.",
    )

    parser.add_argument(
        "-s",
        "--sequence",
        type=parse_sequence_list,
        required=True,
        help="String of comma seperated integer(s) indicating which sequences to run "
        "the pipeline for.",
    )

    parser.add_argument(
        "-a",
        "--algorithm",
        type=str,
        choices=["dtw"],
        required=True,
        help="Pick which algorithm to use. Can be 'dtw'.",
    )

    return parser.parse_args(args)


def main(
    args: Optional[List[str]] = None,
) -> list[dict]:
    """Runs motion tracking orchestrator with command line arguments.

    Args:
         args: A list of command line arguments given as strings. If None, the parser
            will take the args from `sys.argv`.

    Returns:
        A result dict containing saved metrics for specified sequences for all subjects.
    """
    arguments = parse_arguments(args)

    results = orchestrator.run(
        experimental_path=arguments.data,
        gold_path=arguments.gold,
        sequence=arguments.sequence,
        algorithm=arguments.algorithm,
    )

    return results
